import pandas
import random
import re

from jardin.query_builders import \
    SelectQueryBuilder, \
    InsertQueryBuilder, \
    UpdateQueryBuilder, \
    DeleteQueryBuilder, \
    RawQueryBuilder
import jardin.config as config
from jardin.database.database_config import DatabaseConfig


class UnsupportedDriver(Exception): pass


class DatabaseConnections(object):

    _db_configs = {}          # For each db_name, it caches connection info of one or many DB instances

    _connections = {}         # For each db_name, it caches one or many connections

    _active_connections = {}  # For each db_name, it caches only ONE actively used connection. Active connection
                              # can only be changed by a call to self.shuffle_connections(). The idea is to provide
                              # connection stickiness during the lifetime of a session/task/job, etc.

    SUPPORTED_SCHEMES = ('postgres', 'mysql', 'sqlite', 'snowflake', 'redshift')

    @classmethod
    def connection(self, db_name):
        if db_name not in self._active_connections:
            connections = self._connections.get(db_name)
            if connections is None:
                connections = self.build_connections(db_name)
                self._connections[db_name] = connections
            c = connections[0] if len(connections) == 1 else random.choice(connections)
            self.log_db_connection(db_name, c.db_config)
            self._active_connections[db_name] = c
        return self._active_connections[db_name]

    @classmethod
    def build_connections(self, name):
        connections = []
        configs = self.db_configs(name)
        for db in configs:
            if db.scheme not in self.SUPPORTED_SCHEMES:
                raise UnsupportedDriver('%s is not a supported driver' % db.scheme)
            elif db.scheme == 'postgres' or db.scheme == 'redshift':
                import jardin.database.drivers.pg as driver
            elif db.scheme == 'mysql':
                import jardin.database.drivers.mysql as driver
            elif db.scheme == 'sqlite':
                import jardin.database.drivers.sqlite as driver
            elif db.scheme == 'snowflake':
                import jardin.database.drivers.sf as driver
            pool_config = config.CONNECTION_POOLS.get(name, None)
            connections.append(driver.DatabaseConnection(db, name, pool_config=pool_config))
        return connections

    @classmethod
    def db_configs(self, name):
        if len(self._db_configs) == 0:
            config.init()
            for (nme, urls) in config.DATABASES.items():
                if not urls:
                    continue
                # we don't support multi-configs of dictionary format yet; the "else [urls]" is for a dictionary
                url_list = re.split(r'\s+', urls) if isinstance(urls, str) else [urls]
                self._db_configs[nme] = list(map(lambda x: DatabaseConfig(x), url_list))
        return self._db_configs[name]

    @classmethod
    def shuffle_connections(self):
        for name, conns in self._connections.items():
            c = None
            if len(conns) == 1:
                c = conns[0]
            else:
                active = self._active_connections[name]
                filtered = list(filter(lambda x: x is not active, conns))
                c = filtered[0] if len(filtered) == 1 else random.choice(filtered)
            self.log_db_connection(name, c.db_config)
            self._active_connections[name] = c

    @classmethod
    def log_db_connection(self, name, db_config):
        host = getattr(db_config, 'host', None) or '_'  # use "_" for both missing attr or None value cases
        port = getattr(db_config, 'port', None) or '_'
        user = getattr(db_config, 'username', None) or '_'
        database = getattr(db_config, 'database', None) or '_'
        config.logger.debug("[{}]: database connection {}@{}:{}/{}".format(name, user, host, port, database))

def set_defaults(func):
    def wrapper(self, *args, **kwargs):
        kwargs.update(
            model_metadata=self.model_metadata,
            scheme=self.db.db_config.scheme,
            lexicon=self.db.lexicon
            )
        return func(self, *args, **kwargs)
    return wrapper


class DatabaseAdapter(object):

    def __init__(self, db, model_metadata):
        self.db = db
        self.model_metadata = model_metadata

    @set_defaults
    def select(self, **kwargs):
        query = SelectQueryBuilder(**kwargs).query
        config.logger.debug(query)
        results, columns = self.db.execute(*query, write=False)
        if results is None and columns is None:
            return None
        return pandas.DataFrame.from_records(results, columns=columns, coerce_float=True)

    @set_defaults
    def write(self, query_builder, **kwargs):
        query = query_builder(**kwargs).query
        config.logger.debug(query)
        returning_ids = self.db.execute(*query, write=True, **kwargs)
        if len(returning_ids) > 0:
            return self.select(where={kwargs['primary_key']: returning_ids})
        return None

    def insert(self, **kwargs):
        return self.write(InsertQueryBuilder, **kwargs)

    def update(self, **kwargs):
        return self.write(UpdateQueryBuilder, **kwargs)

    @set_defaults
    def delete(self, **kwargs):
        query = DeleteQueryBuilder(**kwargs).query
        config.logger.debug(query)
        self.db.execute(*query, write=False)

    @set_defaults
    def raw_query(self, **kwargs):
        query = RawQueryBuilder(**kwargs).query
        config.logger.debug(query)
        results, columns = self.db.execute(*query, write=False)
        if results is None and columns is None:
            return None
        return pandas.DataFrame.from_records(results, columns=columns, coerce_float=True)


import sys

if not hasattr(sys, 'jardin_setup') or (hasattr(sys, 'jardin_setup') and not sys.jardin_setup):
    from jardin.model import Model, Collection
    from jardin.query import query
    from jardin.transaction import Transaction
    from jardin.tools import reset_session
    import jardin.config

__author__ = 'Emmanuel Turlay'
__copyright__ = 'Copyright 2017, Instacart'
__credits__ = ['Emmanuel Turlay', 'Mathieu Ripert']
__license__ = 'MIT'
__version__ = '0.20.1'
__maintainer__ = 'Emmanuel Turlay'
__email__ = 'emmanuel@instacart.com'
__status__ = 'Prototype'



import sys
import psycopg2 as pg
from psycopg2.pool import SimpleConnectionPool, ThreadedConnectionPool
from psycopg2 import extras
from memoized_property import memoized_property

from jardin.tools import retry
from jardin.database.base import BaseConnection
from jardin.database.lexicon import BaseLexicon
import jardin.config as config


class Lexicon(BaseLexicon):

    @staticmethod
    def table_schema_query(table_name):
        return "SELECT column_name, column_default, data_type FROM " \
            "information_schema.columns WHERE " \
            "table_name=%(table_name)s AND table_schema='public';"

    @staticmethod
    def column_info(row):
        return row['column_name'], row['column_default'], row['data_type']

    @staticmethod
    def update_values(fields, value_extrapolators):
        if len(fields) == 1:
            result = ', '.join(fields) + ' = ' + \
                [', '.join(ext)
                 for ext in value_extrapolators][0]
        else:
            result = '(' \
                + ', '.join(fields) \
                + ') = '
            result += ', '.join(
                ['(' + ', '.join(ext) + ')' for ext in value_extrapolators]
            )
        return result

    @staticmethod
    def row_ids(cursor, primary_key):
        row_ids = cursor.fetchall()
        return [r[primary_key] for r in row_ids]


class DatabaseConnection(BaseConnection):

    DRIVER = pg
    LEXICON = Lexicon

    @retry(pg.OperationalError, tries=3)
    def get_connection(self):
        connection = super(DatabaseConnection, self).get_connection()
        connection.initialize(config.logger)
        return connection

    @memoized_property
    def connect_kwargs(self):
        kwargs = super(DatabaseConnection, self).connect_kwargs
        kwargs.update(connection_factory=extras.MinTimeLoggingConnection)
        return kwargs

    @memoized_property
    def cursor_kwargs(self):
        return dict(cursor_factory=pg.extras.RealDictCursor)

    @memoized_property
    def pool(self):
        if self.pool_config is None:
            return None
        pool_name = self.pool_config.get("pool", None)
        if pool_name is None:
            return None
        pool = getattr(sys.modules[__name__], pool_name)
        min_connections = self.pool_config.get("min_connections", 2)
        max_connections = self.pool_config.get("max_connections", min_connections * 2)
        return pool(min_connections, max_connections, **self.connect_kwargs)

    @retry((pg.InterfaceError, pg.extensions.TransactionRollbackError, pg.extensions.QueryCanceledError), tries=3)
    def execute(self, *query, write=False, **kwargs):
        return super(DatabaseConnection, self).execute(*query, write=write, **kwargs)



from memoized_property import memoized_property
from contextlib import contextmanager
import threading


class BaseConnection(object):

    DRIVER = None
    LEXICON = None

    _connection = None

    def __init__(self, db_config, name, pool_config=None):
        self.db_config = db_config
        self.autocommit = True
        self.name = name
        self.lexicon = self.LEXICON()
        self.pool_config = pool_config

    @contextmanager
    def connection(self):
        try:
            conn = self.get_connection()
            yield conn
            if self.autocommit:
                conn.commit()
        except Exception as e:
            self.rollback()
        finally:
            if self.pool is not None and self.autocommit:
                key = threading.current_thread().ident
                self.pool.putconn(conn, key=key)

    def commit(self):
        conn = self.get_connection()
        conn.commit()
        if self.pool:
            key = threading.current_thread().ident
            self.pool.putconn(conn, key=key)

    def rollback(self):
        conn = self.get_connection()
        conn.rollback()

    @memoized_property
    def connect_kwargs(self):
        return dict(
            database=self.db_config.database,
            user=self.db_config.username,
            password=self.db_config.password,
            host=self.db_config.host,
            port=self.db_config.port,
            connect_timeout=5
        )

    @memoized_property
    def connect_args(self):
        return []

    @memoized_property
    def pool(self):
        return None

    def get_connection(self):
        if self.pool is None:
            if self._connection is None:
                self._connection = self.DRIVER.connect(*self.connect_args, **self.connect_kwargs)
            return self._connection
        key = threading.current_thread().ident
        return self.pool.getconn(key=key)

    @memoized_property
    def cursor_kwargs(self):
        return {}

    def columns(self, cursor):
        cursor_desc = cursor.description
        columns = []
        if cursor_desc:
            columns = [col_desc[0] for col_desc in cursor_desc]
            if self.db_config.lowercase_columns:
                columns = [col.lower() for col in columns]
        return columns

    def execute(self, *query, write=False, **kwargs):
        with self.connection() as connection:
            cursor = connection.cursor(**self.cursor_kwargs)
            cursor.execute(*query)
            if write:
                return self.lexicon.row_ids(cursor, kwargs['primary_key'])
            if cursor.description:
                return cursor.fetchall(), self.columns(cursor)
            return None, None


import re
from memoized_property import memoized_property

import snowflake.connector as sf

from jardin.tools import retry
from jardin.database.drivers.pg import Lexicon as PGLexicon
from jardin.database.base import BaseConnection


class Lexicon(PGLexicon):

    @staticmethod
    def extrapolator(_):
        return '%s'

    @staticmethod
    def format_args(args):
        return args.values()

    @staticmethod
    def standardize_interpolators(sql, params):
        sql, params = super(
            Lexicon, Lexicon).standardize_interpolators(sql, params)
        param_names = re.findall(r'\%\((\w+)\)s', sql)
        if len(param_names):
            sql = re.sub(r'\%\(\w+\)s', '%s', sql)
            if isinstance(params, dict):
                params = list(map(lambda x: params[x], param_names))
        return sql, params


class DatabaseConnection(BaseConnection):

    DRIVER = sf
    LEXICON = Lexicon

    @memoized_property
    def connect_kwargs(self):
        kwargs = dict(
            user=self.db_config.username,
            password=self.db_config.password,
            account=self.db_config.account,
            database=self.db_config.database,
            schema=self.db_config.schema
        )
        if 'warehouse' in dir(self.db_config):
            kwargs['warehouse'] = self.db_config.warehouse
        if 'authenticator' in dir(self.db_config):
            kwargs['authenticator'] = self.db_config.authenticator
        if 'client_session_keep_alive' in dir(self.db_config):
            kwargs['client_session_keep_alive'] = self.db_config.client_session_keep_alive
        return kwargs

    @retry(sf.OperationalError, tries=3)
    def get_connection(self):
        return super(DatabaseConnection, self).get_connection()

    @retry(sf.InterfaceError, tries=3)
    def execute(self, *query, write=False, **kwargs):
        return super(DatabaseConnection, self).execute(*query, write=False, **kwargs)

